@@ -4,7 +4,7 @@ use warnings;
 use 5.010; # state
 
 # ABSTRACT: Fully UTF-8 aware Cwd
-our $VERSION = '0.004'; # VERSION
+our $VERSION = '0.005'; # VERSION
 
 #pod =for test_synopsis
 #pod my $file;
@@ -32,7 +32,7 @@ our $VERSION = '0.004'; # VERSION
 #pod This module replaces all the L<Cwd> functions with fully UTF-8 aware
 #pod versions, both expecting and returning characters.
 #pod
-#pod B<Note:> Replacement of functions is not done on DOS and OS/2
+#pod B<Note:> Replacement of functions is not done on DOS, Windows, and OS/2
 #pod as these systems do not have full UTF-8 file system support.
 #pod
 #pod =head2 Behaviour
@@ -65,38 +65,50 @@ our $VERSION = '0.004'; # VERSION
 #pod
 #pod =cut
 
-use Cwd qw();
-use Encode;
-
-my @EXPORT = qw(cwd getcwd fastcwd fastgetcwd);
-push @EXPORT, qw(getdcwd) if $^O eq 'MSWin32';
-my @EXPORT_OK = qw(chdir abs_path fast_abs_path realpath fast_realpath);
+use Cwd ();
+use Encode ();
 
 # Holds the pointers to the original version of redefined functions
 state %_orig_functions;
 
-# Current package
+# Current (i.e., this) package
 my $current_package = __PACKAGE__;
 
 # Original package (i.e., the one for which this module is replacing the functions)
 my $original_package = $current_package;
 $original_package =~ s/::utf8$//;
 
+require Carp;
+$Carp::Internal{$current_package}++; # To get warnings reported at correct caller level
+
+#pod =attr $File::Find::utf8::UTF8_CHECK
+#pod
+#pod By default C<Cwd:::utf8> marks decoding errors as fatal (default value
+#pod for this setting is C<Encode::FB_CROAK>). If you want, you can change this by
+#pod setting C<Cwd::utf8::UTF8_CHECK>. The value C<Encode::FB_WARN> reports
+#pod the encoding errors as warnings, and C<Encode::FB_DEFAULT> will completely
+#pod ignore them. Please see L<Encode> for details. Note: C<Encode::LEAVE_SRC> is
+#pod I<always> enforced.
+#pod
+#pod =cut
+
+our $UTF8_CHECK = Encode::FB_CROAK; # Die on encoding errors
+
+# UTF-8 Encoding object
+my $_UTF8 = Encode::find_encoding('UTF-8');
+
 sub import {
     # Target package (i.e., the one loading this module)
     my $target_package = caller;
 
-    no strict qw(refs); ## no critic (TestingAndDebugging::ProhibitNoStrict)
-    no warnings qw(redefine);
-
     # If run on the dos/os2/windows platform, ignore overriding functions silently.
-    # These platforms do have (proper) utf-8 file system suppport...
+    # These platforms do not have (proper) utf-8 file system suppport...
     unless ($^O =~ /MSWin32|cygwin|dos|os2/) {
         no strict qw(refs); ## no critic (TestingAndDebugging::ProhibitNoStrict)
         no warnings qw(redefine);
 
         # Redefine each of the functions to their UTF-8 equivalent
-        for my $f (@EXPORT, @EXPORT_OK) {
+        for my $f (@{$original_package . '::EXPORT'}, @{$original_package . '::EXPORT_OK'}) {
             # If we already have the _orig_function, we have redefined the function
             # in an earlier load of this module, so we need not do it again
             unless ($_orig_functions{$f}) {
@@ -107,50 +119,41 @@ sub import {
         $^H{$current_package} = 1; # Set compiler hint that we should use the utf-8 version
     }
 
-    if ($#_) {
-        # Check arguments
-        my @invalid_exports;
-        for my $f (@_[1..$#_]) {
-            if (! grep /^$f$/, (':none', @EXPORT, @EXPORT_OK)) {
-                push @invalid_exports, "$f is not exported by $current_package module";
-            }
-        }
-        if (@invalid_exports) {
-            require Carp;
-            Carp::croak(join("\n", @invalid_exports)  . "\nCan't continue after import errors");
-        }
-    }
+    # Determine symbols to export
+    shift; # First argument contains the package (that's us)
+    @_ = (':DEFAULT') if !@_; # If nothing provided, use default
+    @_ = map { $_ eq ':none' ? () : $_ } @_; # Replace :none tag with empty list
 
-    # Export functions to target package
-    unless ($#_ && grep /^:none$/, @_[1..$#_]) {
-        for my $f ($#_ ? @_[1..$#_] : @EXPORT) {
-            *{$target_package . '::' . $f} = \&{$original_package . '::' . $f};
-        }
-    }
+    # Use exporter to export
+    require Exporter;
+    Exporter::export_to_level($original_package, 1, $target_package, @_) if (@_);
 
     return;
 }
 
-sub unimport {
-    # If run on the dos/os2/windows platform, ignore overriding functions silently.
-    # These platforms do have (proper) utf-8 file system suppport...
-    unless ($^O =~ /MSWin32|cygwin|dos|os2/) {
-        $^H{$current_package} = 0; # Set compiler hint that we should not use the utf-8 version
-    }
-
+sub unimport { ## no critic (Subroutines::ProhibitBuiltinHomonyms)
+    $^H{$current_package} = 0; # Set compiler hint that we should not use the utf-8 version
     return;
 }
 
 sub _utf8_cwd {
     my $func = shift;
+
+    # Enforce LEAVE_SRC
+    $UTF8_CHECK |= Encode::LEAVE_SRC if $UTF8_CHECK;
+
     my $hints = (caller 1)[10]; # Use caller level 1 because of the added anonymous sub around call
     if (! $hints->{$current_package}) {
         # Use original function if we're not using Cwd::utf8 in calling package
         return $_orig_functions{$func}->(@_);
-    } elsif (wantarray) {
-        return map { decode('UTF-8' ,$_) } $_orig_functions{$func}->(map { encode('UTF-8', $_) } @_);
     } else {
-        return decode('UTF-8', $_orig_functions{$func}->(map { encode('UTF-8', $_) } @_));
+        my @args = map { $_ ? $_UTF8->encode($_, $UTF8_CHECK) : $_ } @_;
+        if (wantarray) {
+            return map { $_ ? $_UTF8->decode($_, $UTF8_CHECK) : $_ } $_orig_functions{$func}->(@args);
+        } else {
+            my $r = $_orig_functions{$func}->(@args);
+            return $r ? $_UTF8->decode($r, $UTF8_CHECK) : $r;
+        }
     }
 }
 
@@ -168,7 +171,7 @@ Cwd::utf8 - Fully UTF-8 aware Cwd
 
 =head1 VERSION
 
-version 0.004
+version 0.005
 
 =head1 SYNOPSIS
 
@@ -193,7 +196,7 @@ characters.
 This module replaces all the L<Cwd> functions with fully UTF-8 aware
 versions, both expecting and returning characters.
 
-B<Note:> Replacement of functions is not done on DOS and OS/2
+B<Note:> Replacement of functions is not done on DOS, Windows, and OS/2
 as these systems do not have full UTF-8 file system support.
 
 =head2 Behaviour
@@ -208,6 +211,17 @@ want to prevent this, use C<use Cwd::utf8 qw(:none)>. (As all the
 magic happens in the module's import function, you can not simply use
 C<use Cwd::utf8 qw()>)
 
+=head1 ATTRIBUTES
+
+=head2 $File::Find::utf8::UTF8_CHECK
+
+By default C<Cwd:::utf8> marks decoding errors as fatal (default value
+for this setting is C<Encode::FB_CROAK>). If you want, you can change this by
+setting C<Cwd::utf8::UTF8_CHECK>. The value C<Encode::FB_WARN> reports
+the encoding errors as warnings, and C<Encode::FB_DEFAULT> will completely
+ignore them. Please see L<Encode> for details. Note: C<Encode::LEAVE_SRC> is
+I<always> enforced.
+
 =for test_synopsis my $file;
 
 =head1 BUGS
